// Copyright 2020 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// This starts a mock DAP server that runs indefinitely, accepts DAP
// requests and responds with dummy or error responses.

package main

import (
	"flag"
	"log"
)

func main() {
	port := flag.String("port", "54321", "TCP port to listen on")
	flag.Parse()
	err := server(*port)
	if err != nil {
		log.Fatal("Could not start server: ", err)
	}
}
